/*
 * Copyright (C) 2000-2005 Serge Vakulenko, <vak@cronyx.ru>
 *
 * This file is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You can redistribute this file and/or modify it under the terms of the GNU
 * General Public License (GPL) as published by the Free Software Foundation;
 * either version 2 of the License, or (at your discretion) any later version.
 * See the accompanying file "COPYING.txt" for more details.
 *
 * As a special exception to the GPL, permission is granted for additional
 * uses of the text contained in this file.  See the accompanying file
 * "COPY-UOS.txt" for details.
 */
#include <runtime/lib.h>
#include <kernel/uos.h>
#include <kernel/internal.h>

/*
 * Utility functions.
 * Must be called with interrupts disabled.
 */
void
lock_enqueue (lock_list_t *list, lock_t *elem)
{
	if (! list->head) {
		assert (list->tail == 0);
		elem->next = 0;
		elem->prev = 0;
		list->head = elem;
		list->tail = elem;
	} else {
		assert (list->tail != 0);
		elem->next = list->head;
		elem->prev = 0;
		list->head->prev = elem;
		list->head = elem;
	}
}

void
lock_dequeue (lock_list_t *list, lock_t *elem)
{
	assert (list->head != 0);
	assert (list->tail != 0);
	if (list->head == elem) {
		list->head = elem->next;
		assert (elem->prev == 0);
	} else {
		assert (elem->prev != 0);
		elem->prev->next = elem->next;
	}
	if (list->tail == elem) {
		list->tail = elem->prev;
		assert (elem->next == 0);
	} else {
		assert (elem->next != 0);
		elem->next->prev = elem->prev;
	}
}

/*
 * Get the lock. The calling task would block
 * waiting for the lock to be released by somebody.
 * In the case the lock has associated IRQ number,
 * after acquiring the lock the IRQ will be disabled.
 */
void
lock_take (lock_t *m)
{
	int x;

	MACHDEP_INTR_DISABLE (&x);
	assert (STACK_GUARD (task_current));
	while (m->master && m->master != task_current) {
		/* Monitor is locked, block the task. */
		assert (task_current->lock == 0);
		task_current->lock = m;
		task_dequeue (&task_active, task_current);

		/* Put this task into the list of lock slaves. */
		task_enqueue (&m->slaves, task_current);

		/* Update the value of lock priority.
		 * It must be the maximum of all slave task priorities. */
		if (m->prio < task_current->prio) {
			m->prio = task_current->prio;

			/* Increase the priority of master task. */
			if (m->master->prio < m->prio)
				m->master->prio = m->prio;
		}

		MACHDEP_TASK_SWITCH ();
	}
	if (! m->master) {
		assert (m->slaves.head == 0);
		m->master = task_current;

		/* Put this lock into the list of task slaves. */
		lock_enqueue (&task_current->slaves, m);

		/* Update the value of task priority.
		 * It must be the maximum of base priority,
		 * and all slave lock priorities. */
		if (task_current->prio < m->prio)
			task_current->prio = m->prio;
	}
	MACHDEP_INTR_RESTORE (x);
}

/*
 * Recalculate the value of task priority.
 * It must be the maximum of base priority,
 * and all slave lock priorities.
 */
void
task_recalculate_prio (task_t *t)
{
	lock_t *m;
	int old_prio;

	old_prio = t->prio;
	t->prio = t->base_prio;
	for (m=t->slaves.head; m; m=m->next)
		if (t->prio < m->prio)
			t->prio = m->prio;

	if (t->prio != old_prio && t->lock) {
		if (t->prio > old_prio) {
			/* Priority increased. */
			if (t->lock->prio < t->prio)
				lock_recalculate_prio (t->lock);
		} else {
			/* Priority decreased. */
			if (t->lock->prio <= old_prio)
				lock_recalculate_prio (t->lock);
		}
	}
}

/*
 * Recalculate the value of lock priority.
 * It must be the maximum of all slave task priorities.
 */
void
lock_recalculate_prio (lock_t *m)
{
	task_t *t;
	int old_prio;

	old_prio = m->prio;
	m->prio = 0;
	for (t=m->slaves.head; t; t=t->next)
		if (m->prio < t->prio)
			m->prio = t->prio;

	if (m->prio != old_prio && m->master) {
		if (m->prio > old_prio) {
			/* Priority increased. */
			if (m->master->prio < m->prio)
				task_recalculate_prio (m->master);
		} else {
			/* Priority decreased. */
			if (m->master->prio <= old_prio &&
			    m->master->base_prio < old_prio)
				task_recalculate_prio (m->master);
		}
	}
}

/*
 * Release the lock. All tasks waiting for the lock
 * are unblocked, possibly causing task switch.
 * In the case the lock has associated IRQ number,
 * the IRQ will be enabled.
 */
void
lock_release (lock_t *m)
{
	int x;

	assert (STACK_GUARD (task_current));
	MACHDEP_INTR_DISABLE (&x);
	assert (m->master != 0);

	/* Remove this lock from the list of task slaves. */
	lock_dequeue (&m->master->slaves, m);

	/* Recalculate the value of task priority.
	 * It must be the maximum of base priority,
	 * and all slave lock priorities. */
	if (m->master->prio <= m->prio && m->master->base_prio < m->prio)
		task_recalculate_prio (m->master);
	m->master = 0;

	/* On pending irq, we must call fast handler. */
	if (m->irq) {
		if (m->irq->pending) {
			m->irq->pending = 0;

			/* Unblock all tasks, waiting for irq. */
			if ((m->irq->handler) (m->irq->arg) == 0)
				lock_activate (m, 0);
		}
	}

	if (! m->slaves.head) {
		MACHDEP_INTR_RESTORE (x);
		return;
	}
	do {
		task_t *t = m->slaves.head;
		assert (t->lock == m);
		t->lock = 0;
		task_dequeue (&m->slaves, t);
		task_enqueue (&task_active, t);
	} while (m->slaves.head);
	m->prio = 0;
	MACHDEP_TASK_SWITCH ();
	MACHDEP_INTR_RESTORE (x);
}
